var machina = require( "machina" );
var errs = require( "./customErrors.js" );
var RedisMachine = require( "./redisMachine.js" );
var _ = require( "lodash" );
var when = require( "when" );
var util = require( "util" );
var defaults = {
	enabled: false,
	keyFormat: "",
	redis: {
		host: "127.0.0.1",
		port: 6379,
		options: {},
		ttl: 0
	}
};

var CACHE_DISABLED_DISCONNECT = "Cache disabled, cannot disconnect";
var CACHE_DISABLED_FLUSH = "Cache disabled, cannot flush database";
var CACHE_DISABLED_GET = "Cache disabled, falling back to query for %s";
var CACHE_DISABLED_SET = "Cache disabled, cannot set %s";
var CACHE_DISABLED_SELECT = "Cache disabled, cannot select database %s";
var CACHE_DISABLED_ONREADY = "Cache is disabled.";
var CACHE_INIT = "Cache driver initializing. Config has enabled set to '%s'";
var CACHE_GETTING = "Getting cacheKey: %s";
var CACHE_SETTING = "Setting a %s instance in cache";
var CACHE_SET_COMPLETE = "Cache set, new primary key '%s' for model type %s";
var UNKNOWN_STRATEGY = "Unknown cache strategy configuration: %s";
var INVALID_STRATEGY = "Invalid strategy for type '%s'. %s";

function applyDefaults( config ) {
	return _.merge( {}, defaults, config );
}

function defaultCacheAction( cache, x ) {
	return when( x );
}

function validateStrategy( name, strategy ) {
	var errs = [];
	if ( typeof strategy.getKeyFromParams !== "function" ) {
		errs.push( "Missing getKeyFromParams method" );
	}
	if ( typeof strategy.getKeyFromModel !== "function" ) {
		errs.push( "Missing getKeyFromModel method" );
	}
	if ( typeof strategy.query !== "function" ) {
		errs.push( "Missing query method" );
	}
	if ( errs.length ) {
		throw new InvalidStrategyError( INVALID_STRATEGY, name, errs.join( ". " ) );
	}
}

function setStrategyDefaults( strats ) {
	_.each( strats, function( val, key ) {
		validateStrategy( key, val );
		val.cachify = val.cachify || defaultCacheAction;
		val.deCachify = val.deCachify || defaultCacheAction;
		val.options = val.options || {};
	} );
	return strats;
}

var CacheMachine = machina.Fsm.extend( {
	namespace: "cache.machine",

	initialize: function( options ) {
		var self = this;
		self.strategies = {};
		self.config = applyDefaults( options.config || {} );
		self.logger.debug( CACHE_INIT, self.config.enabled );
		var childFactory = options.redisMachineFactory || ( function() {
			var redisMachine;
			return function redisMachineFactory() {
				redisMachine = redisMachine || new RedisMachine( { config: self.config, logger: self.logger } );
				return redisMachine;
			};
		} )();
		self.states.enabled._child = childFactory;
	},

	initialState: "uninitialized",

	states: {
		uninitialized: {
			_onEnter: function() {
				this.transition( this.config.enabled ? "enabled" : "disabled" );
			}
		},
		enabled: {
			"cache.offline": "disabled"
		},
		disabled: {
			"cache.set": function( key ) {
				this.logger.debug( CACHE_DISABLED_SET, key );
				return when.reject( new CacheDisabledError( CACHE_DISABLED_SET, key ) );
			},
			"cache.get": function( key ) {
				this.logger.debug( CACHE_DISABLED_GET, key );
				return when.reject( new CacheDisabledError( CACHE_DISABLED_GET, key ) );
			},
			"select.db": function( id ) {
				return when.reject( new CacheDisabledError( CACHE_DISABLED_SELECT, id ) );
			},
			"flush.db": function() {
				return when.reject( new CacheDisabledError( CACHE_DISABLED_FLUSH ) );
			},
			disconnect: function() {
				return when.reject( new CacheDisabledError( CACHE_DISABLED_DISCONNECT ) );
			}
		}
	},

	get: function( type, keyParams ) {
		var self = this;
		var strategy = self.strategies[ type ];
		if ( !strategy ) {
			return when.reject( new UnknownStrategyError( UNKNOWN_STRATEGY, type ) );
		}
		var key = self.formattedKey( strategy.getKeyFromParams( keyParams ) );
		var queryOnlyHandler = function queryOnly() {
			return strategy.query( keyParams );
		};
		var queryAndCacheItHandler = function queryAndCacheIt() {
			return strategy.query( keyParams )
				.then( function( queryResult ) {
					return self.set( type, queryResult )
						.then( function() {
							return queryResult;
						} );
				} );
		};
		var deCachifyHandler = function deCache( cacheResult ) {
			return strategy.deCachify( self, cacheResult );
		};
		self.logger.debug( CACHE_GETTING, key );
		return self.handle( "cache.get", key )
			.then( deCachifyHandler )
			.catch( CacheDisabledError, queryOnlyHandler )
			.catch( CacheNotYetConnectedError, queryOnlyHandler )
			.catch( CacheMissError, queryAndCacheItHandler );
	},

	set: function( type, value ) {
		var self = this;
		var strategy = self.strategies[ type ];
		if ( !strategy ) {
			return when.reject( new UnknownStrategyError( UNKNOWN_STRATEGY, type ) );
		}
		self.logger.debug( CACHE_SETTING, type );
		var key = self.formattedKey( strategy.getKeyFromModel( value ) );
		return strategy.cachify( self, _.cloneDeep( value ) )
			.then( function( cacheModel ) {
				function useCachifyResult() {
					return cacheModel;
				}
				return self.handle( "cache.set", key, cacheModel, strategy.options )
					.tap( function() {
						self.logger.debug( CACHE_SET_COMPLETE, key, type );
					} )
					// If the cache has been intentionally disabled
					// or it hasn't established a connection yet...
					// we make it transparent to the caller. Return
					// the cacheModel, and let execution continue.
					.catch( CacheDisabledError, useCachifyResult )
					.catch( CacheNotYetConnectedError, useCachifyResult );
			} );
	},

	loadStrategy: function( opts ) {
		_.merge( this.strategies, setStrategyDefaults( opts ) );
	},

	formattedKey: function( key ) {
		var args = [ key ];
		if ( this.config.keyFormat ) {
			args.unshift( this.config.keyFormat );
		}
		return util.format.apply( null, args );
	},

	flushCache: function() {
		return this.handle( "flush.db" );
	},

	disconnectCache: function() {
		return this.handle( "disconnect" );
	},

	onCacheReady: function() {
		var self = this;
		return when.promise( function( resolve, reject ) {
			if ( self.state === "disabled" ) {
				return reject( new CacheDisabledError( CACHE_DISABLED_ONREADY ) );
			} else if ( self.state === "enabled" &&
					self.states.enabled._child.instance.state === "ready" ) {
				return resolve();
			} else {
				var onReadyHandler = function( data ) {
					if ( data.namespace === "redis.machine" && data.toState === "ready" ) {
						self.off( "transition", onReadyHandler );
						resolve();
					}
				};
				self.on( "transition", onReadyHandler );
			}
		} );
	}
} );

module.exports = CacheMachine;
