var machina = require( "machina" );
var _ = require( "lodash" );
var when = require( "when" );
var redis = require( "redis" );

var CACHE_CONNECTION_ERROR = "Cannot access the cache before the connection to redis has been established";
var CACHE_GET_ERROR = "Cache get Error %s";
var CACHE_HIT = "Cache hit for: %s";
var CACHE_IDLE = "Cache driver client is idle. If the redis server has timeouts enabled, it may disconnect.";
var CACHE_MISS = "Cache miss for: %s";
var CACHE_ONLINE = "Cache driver now online & processing commands";
var CACHE_SET_CONNECTION_WARN = "Attempted to set key %s in the cache before a connection to redis was established";
var CACHE_GET_CONNECTION_WARN = "Attempted to get key %s in the cache before a connection to redis was established";
var CACHE_FLUSH = "Flushed the database";
var CACHE_FLUSH_WARN = "Attempted to flush the database before a connection to redis was established";
var CACHE_FLUSH_ERROR = "Error while attempting to flush the database: %s";
var CACHE_SELECT_WARN = "Attempted to select a database index before a connection to redis was established";
var CACHE_SELECT_ERROR = "Error while attempting to select database index %s: %s";
var CACHE_SELECT = "Selected database index %s";

var redisEvents = [
	{ name: "idle", msg: "Successful", type: "debug" },
	{ name: "connect", msg: "Successful", type: "debug" },
	{ name: "ready", msg: "Ready", type: "debug" },
	{ name: "reconnecting", msg: "Reconnecting", type: "warn" },
	{ name: "error", msg: "Error", type: "error" },
	{ name: "end", msg: "End", type: "debug" }
];

function cacheGetFail( key ) {
	this.logger.warn( CACHE_GET_CONNECTION_WARN, key );
	return when.reject( new CacheNotYetConnectedError( CACHE_CONNECTION_ERROR ) );
}

function cacheSetFail( key ) {
	this.logger.warn( CACHE_SET_CONNECTION_WARN, key );
	return when.reject( new CacheNotYetConnectedError( CACHE_CONNECTION_ERROR ) );
}

function cacheFlushDbFail() {
	this.logger.warn( CACHE_FLUSH_WARN );
	return when.reject( new CacheNotYetConnectedError( CACHE_CONNECTION_ERROR ) );
}

function cacheSelectDbFail() {
	this.logger.warn( CACHE_SELECT_WARN );
	return when.reject( new CacheNotYetConnectedError( CACHE_CONNECTION_ERROR ) );
}

function selectDbIndex( id ) {
	var self = this;
	return when.promise( function( resolve, reject, notify ) {
		self.client.select( id, function( err, result ) {
			if ( err ) {
				self.logger.error( CACHE_SELECT_ERROR, id, err );
				return reject( err );
			}
			self.logger.debug( CACHE_SELECT, id );
			return resolve( id );
		} );
	} );
}

function createRedisClient( config, targetFsm ) {
	var client = redis.createClient( config.port, config.host, config.options );
	redisEvents.forEach( function( evnt ) {
		client.on( evnt.name, function( data ) {
			var args = [ "Redis Connection: %s", evnt.msg ];
			/* istanbul ignore if */
			if ( data ) {
				args.push( data );
			}
			targetFsm.logger[ evnt.type ].apply( targetFsm.logger, args );
			targetFsm.handle( evnt.name );
		} );
	} );
	return client;
}

var RedisMachine = machina.Fsm.extend( {
	namespace: "redis.machine",

	initialState: "initializing",

	initialize: function() {
		this.config.redis.eventsToLog = this.config.redis.eventsToLog || [];
	},

	states: {
		initializing: {
			_onEnter: function() {
				this.client = createRedisClient( this.config.redis, this );
			},
			connect: "connecting",
			"cache.get": cacheGetFail,
			"cache.set": cacheSetFail,
			"flush.db": cacheFlushDbFail,
			"select.db": cacheSelectDbFail,
			"*": function() {
				this.deferUntilTransition();
			}
		},
		probing: {
			_onEnter: function() {
				// check for retry status / error count, etc.
				// may involve knowing context of how we landed here (idle timeout? Error?)
				// but for now:
				this.transition( "connecting" );
			}
		},
		connecting: {
			ready: "selecting",
			error: "probing",
			end: "probing",
			"cache.get": cacheGetFail,
			"cache.set": cacheSetFail,
			"flush.db": cacheFlushDbFail,
			"select.db": cacheSelectDbFail
		},
		selecting: {
			_onEnter: function() {
				var self = this;
				var dbIndex = self.config.redis.dbIndex;
				if ( typeof dbIndex === "number" &&
						this.client.selected_db !== dbIndex ) { //jshint ignore:line
					self.selectDb( self.config.redis.dbIndex )
						.then( function() {
							self.handle( "db.selected" );
						} );
				} else {
					this.transition( "ready" );
				}
			},
			"cache.get": cacheGetFail,
			"cache.set": cacheSetFail,
			"flush.db": cacheFlushDbFail,
			"select.db": selectDbIndex,
			"db.selected": "ready"
		},
		ready: {
			_onEnter: function() {
				if ( this.config.redis.eventsToLog.indexOf( "connected" ) >= 0 ) {
					this.logger.debug( CACHE_ONLINE );
				}
			},
			error: "probing",
			idle: function() {
				if ( this.config.redis.eventsToLog.indexOf( "idle" ) >= 0 ) {
					this.logger.debug( CACHE_IDLE );
				}
			},
			end: "probing",
			disconnect: function() {
				this.client.quit();
				this.transition( "disconnected" );
			},
			"cache.get": function( key ) {
				var self = this;
				return when.promise( function( resolve, reject, notify ) {
					self.client.get( key, function( err, result ) {
						if ( err ) {
							self.logger.error( CACHE_GET_ERROR, err );
							return reject( err );
						}

						if ( _.isNull( result ) ) {
							self.logger.debug( CACHE_MISS, key );
							return reject( new CacheMissError( CACHE_MISS, key ) );
						}

						var parsed;
						try {
							parsed = JSON.parse( result );
						} catch ( e ) {
							parsed = result;
						}
						self.logger.debug( CACHE_HIT, key );
						return resolve( parsed );
					} );
				} );
			},
			"cache.set": function( key, value, options ) {
				var self = this;
				var _options = options || {};
				var ttl = _options.ttl;

				if ( _.isUndefined( ttl ) ) {
					ttl = self.ttl;
				}

				return when.promise( function( resolve, reject, notify ) {
					var _value = value;

					if ( _.isObject( value ) ) {
						_value = JSON.stringify( value );
					}

					self.client.set(
						key,
						_value, function setFn( err, result ) {
							if ( err ) {
								return reject( err );
							}

							if ( ttl && ttl > 0 ) {
								// Set the key expiration
								self.client.expire( key, ttl, function( e ) {
									if ( e ) {
										self.logger.error( "Cache error while setting expiration for key '%s': %s", key, e );
										reject( e );
									}
									self.logger.debug( "Cache set for key '%s'", key );
									return resolve( result );
								} );
							} else {
								self.logger.debug( "Cache set for key '%s'", key );
								return resolve( result );
							}
						}
					);
				} );
			},
			"select.db": selectDbIndex,
			"flush.db": function() {
				var self = this;
				return when.promise( function( resolve, reject, notify ) {
					self.client.flushdb( function( err, result ) {
						if ( err ) {
							self.logger.error( CACHE_FLUSH_ERROR, err );
							return reject( err );
						}
						self.logger.debug( CACHE_FLUSH );
						return resolve( true );
					} );
				} );
			}
		},
		disconnected: {
			_onEnter: function() {
				if ( this.config.redis.eventsToLog.indexOf( "disconnected" ) >= 0 ) {
					this.logger.debug( "Cache driver has disconnected" );
				}
				this.handle( "cache.offline" );
			},
			_reset: "initializing"
		}
	},

	selectDb: function( id ) {
		return this.handle( "select.db", id );
	},

	flushDb: function() {
		return this.handle( "flush.db" );
	}
} );

module.exports = RedisMachine;
